'''Virtual Knots
=============

Class for representing a virtual knot and calling useful methods. This
really subclasses GaussCode, not a space curve.

'''


from pyknot2.representations.gausscode import GaussCode


class
